/*****************************************************************************
 *
 * cddiag.c - show line parameters for a port.
 *
 * Copyright 2000 Digi International (www.digi.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/termios.h>
#include <sys/types.h>
#include <fcntl.h>
#include <linux/serial.h>
#include <sys/errno.h>
#include <linux/fs.h>
#include "fas.h"
#include "ctio.h"
#include "els.h"
#include "cdutils.h"


/*
 *	other protos
 */
int		main(int argc, char **argv);
void	usage(void); 
int		get_line_info(ct_line *lptr, char *devname);
void	show_line_info(ct_line *lp, int show_cmds);

ct_line diagline;

/****************************************************************************
 *
 *	main()
 */
int main(int argc, char **argv) {

		int rc, show_cmds=0;


		if (argc < 2) {
				usage();
				exit(1);
		}
		bzero(&diagline, sizeof(ct_line));
		if ((rc=get_line_info(&diagline, argv[1]))) {
				if (rc != ENOMEM)
						fprintf(stderr, "cddiag: device %s error:%s\n",
								argv[1],strerror(rc));
				else
						fprintf(stderr, "cddiag: device %s hasn't been "
								"opened yet\n", argv[1]);
				exit(0);
		}

		if ((argc > 2) && (strcmp("-c", argv[2])==0))
				show_cmds = 1;

		show_line_info(&diagline, show_cmds);
		return 0; 
}




/****************************************************************************
 *	Do the actual work on the ports 
 *
 */

/*
 * get_line_info()
 * add /dev/ to a portname if needed, and try to open the
 * port. Using the file descriptor from the open,
 * copy the line parameters via the ct_line lptr.
 * return 0 on success, or error code on return.
 *
 */
int get_line_info(ct_line *lptr, char *devname)
{
		char portname[PATH_MAX];
		int diag_fd;
		int rc = 0;
		unsigned int DUTdev=0;
		struct stat DUTstat;
 

		portname[PATH_MAX-1] = 0;
		if (*devname != '/') {
				strcpy(portname, DEV_PATH);
				strncat(portname, devname, PATH_MAX - 5);
		}
		else
				strncpy(portname, devname, PATH_MAX);
 
		if (stat(portname, &DUTstat)) {
				rc = errno;
				fprintf(stderr, "Cannot stat %s:%s\n", devname, strerror(rc));
				return rc;
		}
		DUTdev = (int) DUTstat.st_rdev;
		if (!(S_ISCHR(DUTstat.st_mode))) {
				printf("cddiag: '%s' not a character device\n",portname);
				exit(1);
		}
  
		/* open the diag port */
		diag_fd = open(DEV_DIAG, O_RDWR);
		if (diag_fd < 0) {
				fprintf(stderr, "Couldn't open diagnostic node %s: %s\n",
						DEV_DIAG, strerror(errno));
				exit(1);
		}
		if (diag_fd > 0) {
				lptr->ct_magic = DUTdev;	/* this is how we get DUT devnum
											 * to ioctl. It will break if
											 * ct_magic isn't the first
											 * element in the struct.
											 */
				if (ioctl(diag_fd, STSLINEDIAG, lptr) < 0)
						rc = errno;
				close(diag_fd);
		}
		else
				rc = errno;
		return rc;
}


/*
 * ctf_showfashist()
 *
 * Prints dump of FAS command from history buffer.
 */
LOCAL void ctf_showfashist (u_char *p, int idx)
{
		int front, back;
		char direction;

		front = *p++;
		if (front & 0x80)
				direction = 'R';
		else direction = 'C';
		front &= 0x7f;
		front <<= 8;
		front |= *p++;
		front <<= 8;
		front |= *p++;
		front <<= 8;
		front |= *p++;
		back = *p++;
		back <<= 8;
		back |= *p++;
		back <<= 8;
		back |= *p++;
		back <<= 8;
		back |= *p++;
		if (front|back)
				printf("%2.2d %c %8.8x %8.8x\n", idx, direction, front, back);
}


void show_line_info(struct ct_line *lp, int show_cmds)
{
		int i;
		char on[] = "on ";
		char off[]="off";
		u_char *histptr;

		printf("\n********************** Port Info **********************\n"); 
		
		printf("els%2.2x port %x\n", 0, lp->port);

		printf("Current packet size %d, print_max=%d\n",
			   lp->curr_size_fassend, lp->print_max);
		printf("Baud rate %d  Current baudext %d\n\n",
			   lp->int_obaud,lp->cdbaud_ext); 
		printf("Modem status:\n");
		printf("CTS: %s  DCD: %s DSR: %s  DTR: %s\n",
			   ((lp->mdmsigs & TIOCM_CTS) ? on : off),
			   ((lp->mdmsigs & TIOCM_CAR) ? on : off),
			   ((lp->mdmsigs & TIOCM_DSR) ? on : off),
			   ((lp->mdmsigs & TIOCM_DTR) ? on : off));

		printf("\nFlags:\n");
		printf("offline  : %s       enabled  : %s    disabling : %s \n", 
			   (lp->line_state & LINE_OFFLINE) ? "Y" : "N",
			   (lp->line_state & LINE_ENABLED) ? "Y" : "N",
			   (lp->line_state & LINE_DISABLING) ? "Y" : "N");
		printf("loopback : %s         break  : %s     wait_out : %s\n", 
			   (lp->line_state & LINE_LOOPBACKON) ? "Y" : "N",
			   (lp->line_state & LINE_BREAK) ? "Y" : "N",
			   (lp->line_state & LINE_WAIT_OUT) ? "Y" : "N");
		printf("set_modem: %s     set_params : %s    get_modem : %s\n", 
			   (lp->line_state & LINE_SET_MODEM) ? "Y" : "N",
			   (lp->line_state & LINE_SET_PARAMS) ? "Y" : "N",
			   (lp->line_state & LINE_GET_MODEM) ? "Y" : "N");
		printf("printer  : %s     istopped   : %s      ipending :%s\n", 
			   (lp->line_state & LINE_PRINTER) ? "Y" : "N",
			   (lp->line_state & LINE_ISTOPPED) ? "Y" : "N",
			   (lp->line_state & LINE_IPENDING) ? "Y" : "N");
		printf("obusy :    %s     ostopped   : %s\n", 
			   (lp->line_state & LINE_OBUSY) ? "Y" : "N",
			   (lp->line_state & LINE_OSTOPPED) ? "Y" : "N");
		printf("wait_timer %s     wait_iflush: %s\n", 
			   (lp->line_state & LINE_WAIT_TIMER) ? "Y" : "N",
			   (lp->line_state & LINE_WAIT_IFLUSH) ? "Y" : "N");
		printf("ocsema %d tty %x flags %x\n", lp->ocsema,
			   (unsigned int) lp->tty, (unsigned int) lp->flags);
		printf("refcount %d  openwaitcnt %d\n", lp->refcount, lp->openwaitcnt);
		if (show_cmds && (lp->next_hist >=0) && (lp->next_hist < FASHISTSZ)) {
				printf("\nFAS command history (C=Command sent, "
					   "R=completion recv'd)\n");
				for (i=0 ; i < FASHISTSZ ; i++) {
						histptr =  &lp->fas_hist[FASCMDSZ*((lp->next_hist + i )
														   % FASHISTSZ)];
						ctf_showfashist(histptr, (lp->next_hist+i) %FASHISTSZ);
				}
		}
		/*   disp_termios(lptr); */
		printf("\n");
}


/*
 *  usage()
 *
 *	Display the proper usage and exit
 *
 */
void usage(void) 
{
	fprintf(stderr, "Usage: cddiag devname\n {-c}");
	fprintf(stderr, " where devname is the name of the Etherlite port node\n");
	fprintf(stderr, " and -c enables output of FAS command history.\n");
}


