/*
 *  cdipserv - a utility program with several functions:
 *    1. an alternative to DHCP or BOOTP for assigning IP addresses to
 *    Etherlite units.
 *    2. Reset Etherlites which have become "stuck".
 *    3. Download firmware to Etherlite units.
 *
 *****************************************************************************
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *    Emilio Millan <Emilio_Millan@digi.com>
 *    Mark Schank <Mark_Schank@digi.com>
 *
 *  Authored by Emilio Millan, February 1998, but with this caveat:
 *    Much of the code below was in some way derived from the
 *    source of the CMU bootpd, Berkeley's TFTP server, Mark
 *    Schank's TFTP server, and cdelsreset.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>		/* for malloc() */
#include <string.h>
#include <errno.h>      	/* for errno */
#include <time.h>           	/* for time(), difftime() */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>		/* for socket() */
#include <netdb.h>		/* for getservbyname() */
#include <sys/time.h>		/* for select() */
#include <netinet/in.h>
#include <arpa/inet.h>		/* for inet_ntoa() */
#include <arpa/tftp.h>		/* for struct tftphdr and various constants */
#include <sys/param.h>		/* for MAXHOSTNAMELEN on Irix */


/* used in bp_op in struct bootp */
#define BOOTREPLY     2
#define BOOTREQUEST   1

#define IPPORT_BOOTPS 67  /* port to which bootp binds receive requests */
#define IPPORT_BOOTPC 68  /* the port at which responses are received */
#define IPPORT_TFTP   69  /* trivial FTP, (for firmware update) */

#define MAX_MSG_SIZE    (3*512) /* Maximum packet size */
#define BP_CHADDR_LEN       16
#define BP_SNAME_LEN        64
#define BP_FILE_LEN        128
#define BP_VEND_LEN         64
#define BP_MINPKTSZ        300
#define MAXHADDRLEN          6  /* not really, but for our purposes it'll do */
#define MAX_EL_REQUEST_GAP  60  /* seconds */

#define int32         long
typedef unsigned char HWADDR[6];
typedef unsigned long IPADDR;

struct bootp {
  unsigned char    bp_op;                     /* packet opcode type */
  unsigned char    bp_htype;                  /* hardware addr type */
  unsigned char    bp_hlen;                   /* hardware addr length */
  unsigned char    bp_hops;                   /* gateway hops */
  unsigned int32   bp_xid;                    /* transaction ID */
  unsigned short   bp_secs;                   /* seconds since boot began */
  unsigned short   bp_flags;                  /* RFC1532 broadcast, etc. */
  struct in_addr   bp_ciaddr;                 /* client IP address */
  struct in_addr   bp_yiaddr;                 /* 'your' IP address */
  struct in_addr   bp_siaddr;                 /* server IP address */
  struct in_addr   bp_giaddr;                 /* gateway IP address */
  unsigned char    bp_chaddr[BP_CHADDR_LEN];  /* client hardware address */
  char             bp_sname[BP_SNAME_LEN];    /* server host name */
  char             bp_file[BP_FILE_LEN];      /* boot file name */
  unsigned char    bp_vend[BP_VEND_LEN];      /* vendor-specific area */
  /* note that bp_vend can be longer, extending to end of packet. */
};

#define MAX_AUTH_HOSTS  8
#define NEW_BP_VEND_LEN (MAX_AUTH_HOSTS*10+3+1)

#define USAGE {								\
  printf("usage: %s [flags] hw_addr ip_addr\n", argv[0]);		\
  printf("\n"								\
	 "  flags can include:\n"					\
	 "    -reset    to reset the module before doing anything else\n" \
	 "    -store    to store the IP address in the module's EEPROM\n" \
	 "    -erase    to erase a stored IP from the module's EEPROM\n" \
	 "    -host auth_ip [ auth_netmask ]\n"				\
	 "              to add an IP address or list of IP addresses\n"	\
	 "              to the module's authorized host list\n"		\
	 "              (may be used up to eight times)\n"		\
	 "    -firmware filename\n"					\
	 "              to upload a firmware image to the EtherLite\n"	\
	 "              module once the IP address has been served\n"	\
	 "\n"								\
	 "  hw_addr is the hardware (MAC) address of the device to\n"	\
	 "    be served its IP address.\n"				\
	 "  ip_addr is the IP address to be served.\n"			\
	 "\n"								\
	 "  N.B.: hw_addr should be expressed in hexadecimal\n"		\
	 "        (as on the underside of the EtherLite module),\n"	\
	 "        and ip_addr in Internet standard (decimal) dot\n"	\
	 "        notation (e.g., 192.9.200.10).\n"			\
	);								\
  exit(1);								\
}

/*-------------------------------------------------------------------*/
/* function prototypes */

int str2hwaddr(char *s, HWADDR h);
char *hwaddr2str(HWADDR h);
char *inet_ultoa(unsigned long ul);
char hwaddrs_equal(HWADDR h1, HWADDR h2);
void setarp(int s, struct in_addr *ia, u_char *ha, int len);
void dumpdata(FILE *fp, unsigned char *data, int len);
int reset_elmodule(char *ipaddr);

/*-------------------------------------------------------------------*/
void perror_stdout(char *s) {
	printf("%s: %s\n", s, strerror(errno));
}

#define perror perror_stdout

char *AIX_str_ip = NULL;   /* IP address used in setarp() for AIX */

/*-------------------------------------------------------------------*/
int main(int argc, char *argv[]) {
	int sock_bootp;                /* our socket for BOOTP */
	int sock_tftp;                 /* our socket for TFTP */
	struct servent *servp;         /* used to get values for the next three
					  variables */
	int bootps_port;               /* the port for BOOTP requests */
	int bootpc_port;               /* the port for BOOTP responses */
	int tftp_port;                 /* the port for TFTP requests */
	struct sockaddr_in bind_addr;  /* used to bind to the BOOTPS port */
	struct sockaddr_in recv_addr;  /* the address from which we received
					  the BOOTP request */
	struct sockaddr_in tftp_addr;
	int n;                         /* size of the packet we receive */
	char *pktbuf;                  /* receive packet buffer */
	struct bootp *bp;              /* pktbuf, casted */
	int pktlen;                    /* size of the packet we send back */
	char *str_hwaddr=NULL;         /* hardware address from command line */
	char *str_ipaddr=NULL;         /* IP address from the command line */
	HWADDR hwaddr;                 /* str_hwaddr, parsed */
	IPADDR ipaddr;                 /* str_ipaddr, parsed */
	unsigned char vendor[NEW_BP_VEND_LEN]; /* vendor-specific options */
	int vendor_len=0;              /* length of the above */
	int storing=0, erasing=0, updating=0, resetting=0;
					/* optional program behaviors */
	int num_auth_hosts=0;          /* number of times the -host
					  flag is used */
	int j;                         /* general-purpose integer :-) */
	char *fw_filename;             /* name of firmware file to be uploaded
					  to the module */
	time_t start, now;             /* stuff used to decide when to give
					  a "don't give up" message and when
					  to give a "give up" message */
	int gave_dontgiveup=0, gave_giveup=0;

	/*
	 * Try to grok the command-line arguments.
	 */
	for (j=1; j<argc; j++) {
		if (j==(argc-2)) {str_hwaddr=argv[j]; continue;}
		if (j==(argc-1)) {str_ipaddr=argv[j]; AIX_str_ip=argv[j]; continue;}
		
		/*
		 * -reset 
		 */
		if (strncmp("-r", argv[j], 2)==0) {
			resetting=1;
#     ifdef PRINT_ARG_PARSE
			printf("Will reset the module.\n");
#     endif
			continue;
		}
		
		/*
		 * -store
		 */
		if (strncmp("-s", argv[j], 2)==0) {
			if (erasing) {
				printf("You cannot use -store and -erase "
				       "simultaneously.\n\n");
				USAGE;
			}
			
			vendor[vendor_len++]=3;
			vendor[vendor_len++]=1;
			vendor[vendor_len++]=1;
			storing=1;
#     ifdef PRINT_ARG_PARSE
			printf("Will store the IP address.\n");
#     endif
			continue;
		}

		/*
		 * -erase
		 */
		if (strncmp("-e", argv[j], 2)==0) {
			if (storing) {
				printf("You cannot use -store and -erase "
				       "simultaneously.\n\n");
				USAGE;
			}
			
			vendor[vendor_len++]=3;
			vendor[vendor_len++]=1;
			vendor[vendor_len++]=2;
			erasing=1;
#     ifdef PRINT_ARG_PARSE
			printf("Will erase the stored IP address.\n");
#     endif
			continue;
		}

		/*
		 * -host ip_addr [ netmask ]
		 */
		if (strncmp("-h", argv[j], 2)==0) {
			if (num_auth_hosts==MAX_AUTH_HOSTS) {
				printf("Only %d hosts may be declared "
				       "authorized.\n", MAX_AUTH_HOSTS);
				printf("Consider using host lists instead of "
				       "specifying individual hosts.\n\n");
			  USAGE;
			}

			if (j==(argc-3) || argv[j+1][0]=='-') {
				/* Bad use of -hosts. */
				printf("-hosts requires an IP address (and "
				       "optional bitmask)\n\n");
				USAGE;
			}
			if (j==(argc-4) || argv[j+2][0]=='-') {
				/* This is a one-arg use of -hosts. */
				
				IPADDR auth_ip;

				auth_ip=inet_addr(argv[j+1]);
				if (auth_ip==(-1)) {
					printf("Bad IP address: \"%s\"\n\n",
					       argv[j+1]);
					USAGE;
				}
				
				auth_ip=htonl(auth_ip);

				vendor[vendor_len++]=1;
				vendor[vendor_len++]=4;
				vendor[vendor_len++]=(auth_ip&0xFF000000) >>24;
				vendor[vendor_len++]=(auth_ip&0x00FF0000) >>16;
				vendor[vendor_len++]=(auth_ip&0x0000FF00) >> 8;
				vendor[vendor_len++]=(auth_ip&0x000000FF) >> 0;

				j++;
				num_auth_hosts++;
#       ifdef PRINT_ARG_PARSE
				printf("Adding IP address %s to the authorized"
				       " host list.\n", inet_ultoa(auth_ip));
#       endif
				continue;
		
			} else {
				/* This is a two-arg use of -hosts. */

				IPADDR auth_ip, auth_netmask;

				auth_ip=inet_addr(argv[j+1]);
				if (auth_ip==(-1)) {
					printf("Bad IP address: \"%s\"\n\n",
					       argv[j+1]);
					USAGE;
				}

				auth_ip=htonl(auth_ip);
				
				auth_netmask=inet_addr(argv[j+2]);
				if (auth_netmask==(-1)) {
					printf("Bad netmask: \"%s\"\n\n",
					       argv[j+2]);
					USAGE;
				}
				
				auth_netmask=htonl(auth_netmask);

				vendor[vendor_len++]=2;
				vendor[vendor_len++]=8;
				vendor[vendor_len++]=
					(auth_netmask&0xFF000000) >> 24;
				vendor[vendor_len++]=
					(auth_netmask&0x00FF0000) >> 16;
				vendor[vendor_len++]=
					(auth_netmask&0x0000FF00) >>  8;
				vendor[vendor_len++]=
					(auth_netmask&0x000000FF) >>  0;
				vendor[vendor_len++]=
					(auth_ip&0xFF000000) >> 24;
				vendor[vendor_len++]=
					(auth_ip&0x00FF0000) >> 16;
				vendor[vendor_len++]=
					(auth_ip&0x0000FF00) >>  8;
				vendor[vendor_len++]=
					(auth_ip&0x000000FF) >>  0;

				dumpdata(stdout, vendor, vendor_len);
		
				j+=2;
				num_auth_hosts++;
		
#       ifdef PRINT_ARG_PARSE
				printf("Adding IP address %s/",
				       inet_ultoa(auth_ip));
				printf("netmask %s to the authorized host "
				       "list.\n", inet_ultoa(auth_netmask));
#       endif
				continue;
			}
		} /* -hosts */

		/*
		 * -file (or -firmware)
		 */
		if (strncmp("-f", argv[j], 2)==0) {
			struct stat statbuf;
			
			if (j==(argc-3) || argv[j+1][0]=='-') {
				printf("-file (-firmware) requires an "
				       "argument: the file to be uploaded"
				       "\n\n");
				USAGE;
			}
			if (updating) {
				printf("Only one file may be uploaded.\n\n");
				USAGE;
			}
			
			if (strlen(argv[j+1]) >= (BP_FILE_LEN-1)) {
				printf("Filename is too long: \"%s\"."
				       "Must be %d characters or fewer.\n\n",
				       argv[j+1], BP_FILE_LEN-1);
				USAGE;
			}

			if (stat(argv[j+1], &statbuf)<0) {
				printf("Can't open file \"%s\".\n", argv[j+1]);
				perror("stat");
				printf("\n");
			  USAGE;
			}
		
			updating=1;
			fw_filename=argv[j+1];
			j++;
			
#     ifdef PRINT_ARG_PARSE
			printf("Will send file \"%s\" to the module.\n",
			       fw_filename);
#     endif
			continue;
		}
	
		printf("Unrecognized argument: \"%s\"\n\n", argv[j]);
		USAGE;
	} /* looping through argv[] */
  
	/*
	 * Got enough args?
	 */
	if (str_hwaddr==NULL || str_ipaddr==NULL) {
		printf("Too few arguments.\n\n");
		USAGE;
	}

	/*
	 * Make sure the HW address is kosher.
	 */
	/* Replace dashes with colons--otherwise sscanf() pukes. */
	for (j=0; j<strlen(str_hwaddr); j++)
		if (str_hwaddr[j]=='-')
			str_hwaddr[j]=':';
	if (str2hwaddr(str_hwaddr, hwaddr)<0) {
		printf("Bad hardware address: \"%s\"\n", str_hwaddr);
		printf("A hardware address should be a six-byte value "
		       "(e.g., 00-A0-E7-00-03-63).\n\n");
		USAGE;
	}

	/*
	 * Make sure the IP address is kosher.
	 */
	ipaddr=inet_addr(str_ipaddr);
	if (ipaddr==(-1)) {
		printf("Bad IP address: \"%s\"\n", str_ipaddr);
		printf("An IP address should be a four-byte value "
		       "(e.g., 192.9.200.10).\n\n");
		USAGE;
	}

# ifdef PRINT_ARG_PARSE
	printf("hardware address: %s\n", hwaddr2str(hwaddr));
	printf("IP address:       %s\n", inet_ultoa(ipaddr));
	printf("\n");
# endif
	/*
	 * Get space for receiving packets and composing replies.
	 */
	pktbuf = malloc(MAX_MSG_SIZE);
	if (!pktbuf) {
		perror("malloc");
		return(1);
	}
	bp = (struct bootp *) pktbuf;

	/*
	 * Open up a socket.
	 */
	if ((sock_bootp = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	/*
	 * Get server's listening port number.
	 */
	servp = getservbyname("bootps", "udp");
	if (servp) {
		bootps_port = ntohs((u_short) servp->s_port);
	} else {
		bootps_port = (u_short) IPPORT_BOOTPS;
		printf("(udp/bootps: unknown service--assuming port %d)\n",
		       bootps_port);
	}

	/*
	 * Bind socket to BOOTPS port.
	 */
	bind_addr.sin_family = AF_INET;
	bind_addr.sin_addr.s_addr = INADDR_ANY;
	bind_addr.sin_port = htons(bootps_port);
	if (bind(sock_bootp, (struct sockaddr *) &bind_addr,
		 sizeof(bind_addr)) < 0) {
		perror("bind");
		if (errno==EADDRINUSE) {
			printf("Most likely there is a BOOTP server "
			       "already running on this system.\n");
		}
		if (errno==EACCES) {
			printf("You will need root access "
			       "to run this command on this system.\n");
		}
		exit(1);
	}

	/*
	 * Get destination port number so we can reply to client.
	 */
	servp = getservbyname("bootpc", "udp");
	if (servp) {
	bootpc_port = ntohs(servp->s_port);
	} else {
		bootpc_port = (u_short) IPPORT_BOOTPC;
		printf("(udp/bootpc: unknown service--assuming port %d)\n",
		       bootpc_port);
	}

	if (updating) {
		/*
		 * Get TFTP port.
		 */
		servp = getservbyname("tftp", "udp");
		if (servp) {
			tftp_port = ntohs(servp->s_port);
		} else {
			tftp_port = (u_short) IPPORT_TFTP;
			printf("(udp/tftp: unknown service--assuming "
			       "port %d)\n", tftp_port);
		}

		if ((sock_tftp=socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			perror("socket");
			exit(1);
		}
	
		tftp_addr.sin_family = AF_INET;
		tftp_addr.sin_addr.s_addr = INADDR_ANY;
		tftp_addr.sin_port = htons(tftp_port);
		if (bind(sock_tftp, (struct sockaddr *) &tftp_addr,
			 sizeof(struct sockaddr_in)) < 0) {
			perror("bind");
			if (errno==EADDRINUSE) {
				printf("Most likely there is a TFTP server "
				       "already running on this system.\n");
			}
			if (errno==EACCES) {
				printf("You will need root access to "
				       "run this command on this system.\n");
			}
			exit(1);
		}
	} /* if (updating) */


	if (resetting) {
		printf("Resetting the EtherLite module...\n");
		if (reset_elmodule(str_ipaddr)<0)
			printf("Remote reset failed.  You will have to "
			       "manually power cycle the module.\n");
	}

	/*
	 * Process incoming requests.
	 */
	printf("Awaiting a BOOTP request from HW address %s...\n",
	       hwaddr2str(hwaddr));
	printf("(This may take up to a minute.)\n");

	start=time(NULL);
	for (;;) {
		fd_set readset;
		struct timeval timeout={5,0}; /* every 5 seconds */
		int ra_len;
		
		FD_ZERO(&readset);
		FD_SET(sock_bootp, &readset);
		
		/*
		 * Time to emit some words of wisdom?
		 */
		now=time(NULL);
		if (!gave_giveup && difftime(now, start) > 
		    MAX_EL_REQUEST_GAP + timeout.tv_sec) {
			printf("If you are trying to serve an EtherLite "
			       "module,\n"
			       "a BOOTP request should have been "
			       "received by now.\n"
			       "A problem may exist.\n");
			gave_giveup=1;
		}
		
		if (!gave_dontgiveup && difftime(now, start) >=
		    (MAX_EL_REQUEST_GAP/2)) {
			printf("Don't give up hope!\n");
			gave_dontgiveup=1;
		}

		/*
		 * Wait for something to come in...
		 */
		if (select(sock_bootp+1, &readset, NULL, NULL, &timeout)!=1) {
			printf("None yet...\n");
			continue;
		}

		/*
		 * Something came in, so get it.
		 */
		ra_len = sizeof(recv_addr);
		n = recvfrom(sock_bootp, pktbuf, MAX_MSG_SIZE, 0,
			     (struct sockaddr *)&recv_addr, &ra_len);

		/*
		 * Check for bad things.
		 */
		if (n <= 0) {
			/* Uh-oh. */
			perror("recvfrom");
			continue;
		}

		if (n < sizeof(struct bootp)) {
			printf("(Ignoring a short packet.)");
			continue;
		}
		pktlen=n;

		if (bp->bp_op!=BOOTREQUEST) {
			/* printf("Not a BOOTREQUEST.\n"); */
			continue;
		}

		if (bp->bp_hlen!=6) {
			printf("Bad hardware-address length (%d)\n",
			       bp->bp_hlen);
			continue;
		}

		/*
		 * All looks good so far.  Make sure the request is for us.
		 */
		if (!hwaddrs_equal(hwaddr, bp->bp_chaddr)) {
			printf("(Ignoring a BOOTP request from %s.)\n",
			       hwaddr2str(bp->bp_chaddr));
			continue;
		}

		/*
		 * OK--let's rock and roll!
		 */
		printf("Got one--responding...\n");
		break;
	}

# ifdef DEBUG
	printf("recvd pkt from IP addr %s\n", inet_ntoa(recv_addr.sin_addr));
	printf("bp_hlen=%d\n", bp->bp_hlen);
	printf("addr=%02X.%02X.%02X.%02X.%02X.%02X\n",
	       bp->bp_chaddr[0], bp->bp_chaddr[1], bp->bp_chaddr[2],
	       bp->bp_chaddr[3], bp->bp_chaddr[4], bp->bp_chaddr[5]);
# endif

	/*
	 *  Let's reply!
	 */
	{
		struct in_addr dst;
		u_short port;
		struct sockaddr_in send_addr;
		
		/*
		 * Turn this into a reply.
		 */
		bp->bp_op=BOOTREPLY;
		
		/*
		 * This is the address we're handing back.
		 */
		(bp->bp_yiaddr).s_addr=ipaddr;
  	
		if (bp->bp_ciaddr.s_addr) {
			/* If the request came it with an IP address, send the
			 * response to that IP address.  With EtherLites,
			 * this'll never happen.
			 */
			dst = bp->bp_ciaddr;
			port = bootpc_port;
		} else if (bp->bp_giaddr.s_addr) {
			/*
			 * Otherwise, if it came in via a gateway, send
			 * the response to the gateway.
			 */
			printf("(Responding via the gateway at %s...)\n",
			       inet_ntoa(bp->bp_giaddr));
			dst = bp->bp_giaddr;
			port = bootps_port;
		} else {
			unsigned char *ha;
			int len;

			/*
			 * Otherwise, send it back using the hardware address
			 * we got as the address by doing so ARP trickery.
			 */
			dst = bp->bp_yiaddr;
			port = bootpc_port;
			
			ha = bp->bp_chaddr;
			len = bp->bp_hlen;
			if (len > MAXHADDRLEN)
				len = MAXHADDRLEN;
			setarp(sock_bootp, &dst, ha, len);
		}

		/* Set up socket address for send. */
		send_addr.sin_family = AF_INET;
		send_addr.sin_port = htons(port);
		send_addr.sin_addr = dst;
	
		if (connect(sock_bootp, (struct sockaddr *)&send_addr,
			    sizeof(send_addr)) < 0) {
			perror("connect");
		}

		/*
		 * Tack on the vendor-specific stuff.
		 */
		if (vendor_len==0) {
			bp->bp_vend[4]=0xFF;
			pktlen=sizeof(struct bootp) - BP_VEND_LEN + 4 + 1;
		} else {
			/* Tack on the needed bytes after the magic cookie. */
			bp->bp_vend[4]=43;	       /* vendor specific */
			bp->bp_vend[5]=vendor_len;   /* bytes that follow */
			memcpy(&bp->bp_vend[6], vendor, vendor_len);
			bp->bp_vend[6+vendor_len]=0xFF;
			pktlen=sizeof(struct bootp)
				- BP_VEND_LEN  /* We'll count it ourself. */
				+ vendor_len  /* Here is is. */
				+ 4  /* ...plus the magic cookie */
				+ 2  /* ...plus the 43 and the length byte */
				+ 1; /* ...plus "stop scanning" token (0xFF) */
		}

#   ifdef DEBUG
		{
			int i;
			for (i=0; bp->bp_vend[i]!=0xFF; i++)
				printf("%d: %02X\n", i, bp->bp_vend[i]);
		}

		printf("sizeof(struct bootp)=%d pktlen=%d\n",
		       sizeof(struct bootp),
		       pktlen);
#   endif

		/*
		 * Figure out our IP address.
		 */
#define USE_HOSTNAME_TO_GET_IPADDR
#ifdef USE_HOSTNAME_TO_GET_IPADDR
		{
			char hostname[MAXHOSTNAMELEN+1];
			struct hostent *he;
			struct in_addr my_ip_addr;
			
			if (gethostname(hostname, MAXHOSTNAMELEN)==0) {
				if ((he=gethostbyname(hostname))!=NULL) {
					memcpy(&my_ip_addr, he->h_addr,
					       sizeof(my_ip_addr));
					
					/* This is us. */
					bp->bp_siaddr=my_ip_addr;
				}
			}
		}
#else
		{
			struct sockaddr_in sa;
			int sa_len;
			
			/* This *should* work, but... */

			sa_len=sizeof(sa);
			if (getsockname(sock_bootp, (struct sockaddr *)&sa,
					&sa_len) < 0)
				perror("getsockname");
			else
				bp->bp_siaddr=sa.sin_addr;
			printf("%s\n", inet_ntoa(sa.sin_addr));
		}
#endif

		if (updating) {
			memcpy(bp->bp_file, fw_filename,strlen(fw_filename)+1);
		}

		/* Send reply with same size packet as request used. */
		if (send(sock_bootp, pktbuf, pktlen, 0) < 0) {
			perror("send");
			exit(1);
		}
	}

	/*
	 * Print bombastic warnings as necessary.
	 */
	
	if (erasing) {
		printf(
"  +------------------------------------------------------------------+\n"
"  | If the EtherLite module did NOT have a stored IP address, the    |\n"
"  | \"on\" indicator will soon be steadily lit, and the module will    |\n"
"  | then be ready for use.                                           |\n"
"  |                                                                  |\n"
"  | If the EtherLite module DID have a stored IP address, then it    |\n"
"  | will now reprogram itself to erase that IP address.  This will   |\n"
"  | take about 10 seconds.  DO NOT DISRUPT POWER TO THE MODULE UNTIL |\n"
"  | THE \"ON\" INDICATOR HAS RESUMED RAPID FLASHING FOR A FULL 5     |\n"
"  | SECONDS.  Failure to do so could render the EtherLite module     |\n"
"  | inoperable.                                                      |\n"
"  |      Once reprogramming has finished, the EtherLite module will  |\n"
"  | await delivery of a new IP address.  You may use this program    |\n"
"  | to deliver this address.                                         |\n"
"  +------------------------------------------------------------------+\n");
	} else {
		if (storing) {
			printf(
"  +---------------------------------------------------------------------+\n"
"  | If the EtherLite module did not have a stored IP address, or if the |\n"
"  | address to be stored is different than the one currently stored,    |\n"
"  | the EtherLite module is now reprogramming itself.  This will take   |\n"
"  | about 10 seconds.  DO NOT DISRUPT POWER TO THE MODULE UNTIL THE     |\n"
"  | \"ON\" INDICATOR HAS REMAINED STEADILY LIT FOR A FULL 5 SECONDS.      |\n"
"  | Failure to do so could render the EtherLite module inoperable.      |\n"
"  | Once reprogramming has finished, the EtherLite module will be ready |\n"
"  | for use.                                                            |\n"
"  +---------------------------------------------------------------------+\n");
		}
		printf("Address successfully served.\n");
	}

	if (updating) {
		char pktbuf[sizeof(struct tftphdr)+SEGSIZE];
		int pktlen=sizeof(struct tftphdr)+SEGSIZE;
		struct tftphdr *t=(struct tftphdr *)pktbuf;
		int recv_cnt;
		struct sockaddr_in from;
		size_t addrlen=sizeof(struct sockaddr_in);
		int fd;
		int last_nread=999;

		printf("Awaiting a TFTP request...\n");

		for (;;) {
			fd_set readset;
			struct timeval timeout={5,0}; /* every 5 seconds */

			FD_ZERO(&readset);
			FD_SET(sock_tftp, &readset);

			if (select(sock_tftp+1, &readset, NULL, NULL,
				   &timeout) != 1) {
				printf("The module has been quiescent for "
				       "%d seconds.\n"
				       "Assuming something is wrong...\n",
				       (int)timeout.tv_sec);
				goto bail;
			}

			if ((recv_cnt=recvfrom(sock_tftp, pktbuf, pktlen, 0,
					       (struct sockaddr *)&from,
					       &addrlen)) < 0) {
				perror("recvfrom");
				continue;
			}
	
			switch (ntohs(t->th_opcode)) {
			case RRQ:
				/*
				 * This is a read request--the module is
				 * asking for its firmware.
				 */

#ifdef DEBUG
				printf("The module wants file %s.\n",
				       (char *)&t->th_block);
#endif
		
				/*
				 * Open up the file for sending.
				 */
				fd=open((char *)&t->th_block, O_RDONLY);
				if (fd<0) {
					perror("open");
					goto bail;
				}

				/*
				 * Ready the first data packet to go out.
				 */
				t->th_opcode=htons(DATA);
				t->th_block=htons(1); /* Yep, first one's
						       * numbered 1, not 0.
						       * Guess who made that
						       * mistake.
						       */
				if (read(fd, &t->th_data, SEGSIZE)<0) {
					perror("read");
					goto bail;
				}

				/*
				 * We use the same socket for receives 
				 * as for sends.
				 */
				if (connect(sock_tftp,
					    (struct sockaddr *)&from,
					    sizeof(from)) < 0) {
					perror("connect");
					exit(1);
				}
		
				printf("Starting to upload the firmware...\n");

				/* Send her off! */
				if (send(sock_tftp, pktbuf, SEGSIZE+4, 0) !=
				    SEGSIZE+4) {
					perror("send");
					goto bail;
				}
				/* Await the module's response. */
				break;

			case ACK: {
				/*
				 * The module firmware has acknowledged 
				 * the data block we sent.
				 */
				int nread;

				/* Get the next hunk of data. */
				if (lseek(fd, ntohs(t->th_block)*SEGSIZE,
					  SEEK_SET) < 0)
					goto bail;

				/* Read the packet to go. */
				t->th_opcode=htons(DATA);
				t->th_block=htons(ntohs(t->th_block)+1);
				nread=read(fd, &t->th_data, SEGSIZE);

				if (nread<0) {
					perror("read");
					goto bail;
				}
		
				/* Hold on--are we done? */
				if (nread==0 && last_nread==0) {
					/* Yes--hooray! */
					printf("Firmware successfully "
					       "uploaded.\n");
					goto bail;
				}
				last_nread=nread;

				/* Nope--packet away! */
				if (send(sock_tftp, pktbuf, nread+4, 0) != 
				    nread+4) {
					perror("send");
					goto bail;
				}
				break;
			}

			case ERROR:
				printf("The module reported a TFTP error.\n");
				goto bail;

			case WRQ:
			case DATA:
			default:
				printf("Received an unexpected TFTP request "
				       "(%d).\n", ntohs(t->th_opcode));
				goto bail;

			} /* switch (opcode) */
		} /* for loop */
	bail:
		close(sock_tftp);
	} /* if (updating) */


	return(0);  /* That's it! */
}



/*-------------------------------------------------------------------*/
int str2hwaddr(char *s, HWADDR h) {
	unsigned int ch[6];
	int i;

	if (sscanf(s, "%2x%2x%2x%2x%2x%2x",
		   &ch[0], &ch[1], &ch[2], &ch[3], &ch[4], &ch[5])==6);
	else if (sscanf(s, "%2x.%2x.%2x.%2x.%2x.%2x",
			&ch[0], &ch[1], &ch[2], &ch[3], &ch[4], &ch[5])==6);
	else if (sscanf(s, "%2x:%2x:%2x:%2x:%2x:%2x",
			&ch[0], &ch[1], &ch[2], &ch[3], &ch[4], &ch[5])==6);
	else
		return(-1);

	for (i=0; i<6; i++) {
		((char *)h)[i]=ch[i];
	}

#ifdef DEBUG
	printf("Address parsed as %02X:%02X:%02X:%02X:%02X:%02X\n",
	       ch[0], ch[1], ch[2], ch[3], ch[4], ch[5]);
#endif
	return(0);
}


/*-------------------------------------------------------------------*/
char *hwaddr2str(HWADDR h) {
	static char s[64];
	unsigned char *ch=h;

	sprintf(s, "%02X:%02X:%02X:%02X:%02X:%02X",
		ch[0], ch[1], ch[2], ch[3], ch[4], ch[5]);
	return(s);
}


/*-------------------------------------------------------------------*/
char hwaddrs_equal(HWADDR h1, HWADDR h2) {
	unsigned char *ch1=h1, *ch2=h2;

	return(ch1[0]==ch2[0] && ch1[1]==ch2[1] &&
	       ch1[2]==ch2[2] && ch1[3]==ch2[3] &&
	       ch1[4]==ch2[4] && ch1[5]==ch2[5]);
}


/*-------------------------------------------------------------------*/
char *inet_ultoa(unsigned long ul) {
	struct in_addr a;

	a.s_addr=ul;
	return(inet_ntoa(a));
}


/*-------------------------------------------------------------------*/
#include <sys/ioctl.h>
/*  else */
/*  # include <sys/sockio.h> */
/*  #endif */
#include <net/if_arp.h>
#define bcopy(a,b,c)    memcpy(b,a,c)
#define bzero(p,l)      memset(p,0,l)

/*
 * Setup the arp cache so that IP address 'ia' will be temporarily
 * bound to hardware address 'ha' of length 'len'.
 */
void setarp(int s, struct in_addr *ia, u_char *ha, int len) {

	struct arpreq arpreq;       /* Arp request ioctl block */
	struct sockaddr_in *si;

#ifdef DEBUG
	printf("IP address is %s\n", inet_ntoa(*ia));
#endif

	bzero((caddr_t) & arpreq, sizeof(arpreq));
	arpreq.arp_flags = ATF_COM;

	/* Set up the protocol address. */
	arpreq.arp_pa.sa_family = AF_INET;
	si = (struct sockaddr_in *) &arpreq.arp_pa;
	si->sin_addr = *ia;

	/* Set up the hardware address. */
	bcopy(ha, arpreq.arp_ha.sa_data, len);
	arpreq.arp_ha.sa_family = ARPHRD_ETHER;
	
#if defined(AIX4)
	/*
	 * use command line service to set arp table, since ioctl
	 * (SIOCSARP) fails
	 */
	{ int rc;
	char sstring[256];

	printf("HW = %s\n", hwaddr2str(ha));
	printf("IP = %s \n", AIX_str_ip);
	sprintf(sstring, "/usr/sbin/arp -s ether %s %s", AIX_str_ip,
                hwaddr2str(ha));

	rc = system(sstring);

	if (rc) {
		perror("system(arp)");
		printf("rc = %#02x\n", rc);
		printf("update arp table failed\n");
	}
	/*
	  arpreq.arp_pa.sa_len = 4;
	  bcopy(ia, arpreq.arp_pa.sa_data, 4);

	  arpreq.at_length = len;
	*/
	}
#else /* AIX4 */

	/*
	 * On SunOS, the ioctl sometimes returns ENXIO, and it
	 * appears to happen when the ARP cache entry you tried
	 * to add is already in the cache.  (Sigh...)
	 * XXX - Should this error simply be ignored? -gwr
	 */
	if (ioctl(s, SIOCSARP, (caddr_t) &arpreq) < 0) {
		perror("ioctl(SIOCSARP)");
      printf("NOTE: The operating system reported a problem when cdipserv\n"
	     "attempted to prepare the ARP table for sending the BOOTP\n"
	     "response.  As such, the response may or may not reach the\n"
	     "module.  After cdipserv claims to have sent off its response,\n"
	     "you should check to see if the module received its IP address.\n"
	     "If not, reboot this host and try again.\n");
	}
#endif /* AIX4 */
}


/*---------------------------------------------------------------*/
#include <ctype.h>

void dumpdata(FILE *fp, unsigned char *data, int len) {
	int i, j;
	char tmp[40];

	for (i=0; i<len; i+=16) {
		sprintf(tmp, "%5d/0x%X: ", i, i);
		fprintf(fp, "%12s", tmp);
		for (j=0; j<16; j++)
			if (i+j<len)
				fprintf(fp, "%02X ", data[i+j]);
			else
				fprintf(fp, "   ");
		fprintf(fp, " ");
		for (j=0; j<16; j++)
      if (i+j<len) {
	      if (isprint((char)data[i+j]))
		      fprintf(fp, "%c", data[i+j]);
	      else
		      fprintf(fp, ".");
      }
		fprintf(fp, "\n");
	}
}


/*---------------------------------------------------------------*/
#include "fas.h"

int sock_read(int s, caddr_t buf, int wanted);
void bitmash(unsigned char *challenge, unsigned char *key);

#define ELS_TCP_PORT    10001
#define FASCMDSZ        8

int reset_elmodule(char *ipaddr) {
	int s;
	struct sockaddr_in el_addr;
	req_unit_unlock	unlock_req;
	rsp_unit_unlock	unlock_rsp;
	req_unit_inquiry	inquiry_req;
	rsp_unit_inquiry	inquiry_rsp;
	req_global		reset_req;
	rsp_global		reset_rsp;
	char              str_inquiry[64];

	/*
	 * Create the socket and the IP connect to the module.
	 */
	if ((s=socket(AF_INET, SOCK_STREAM, 0))<0) {
		perror("socket");
		return(-1);
	}

	el_addr.sin_family = AF_INET;
	el_addr.sin_addr.s_addr = inet_addr(ipaddr);
	el_addr.sin_port = htons(ELS_TCP_PORT);
	
	if (connect(s, (struct sockaddr *)&el_addr, sizeof(el_addr)) < 0) {
		perror("connect");
		close(s);
		return(-1);
	}

	/*
	 * Request a challenge from the ELS.
	 */

	unlock_req.opcode = FAS_UNIT_UNLOCK;
	if (write(s, &unlock_req, FASCMDSZ) < 0) {
		perror("write");
		close(s);
		return(-1);
	}

	if (sock_read(s, (caddr_t)&unlock_rsp, FASCMDSZ) < 0) {
		perror("read");
		close(s);
		return(-1);
	}

	if (unlock_rsp.opcode != FAS_UNIT_UNLOCK) {
		printf("Bad unlock response received from "
		       "EtherLite module.\n");
		close(s);
		return(-1);
	}

	/*
	 * Do the math that'll prove to the module that we're kosher.
	 */
	bitmash(unlock_rsp.challenge,inquiry_req.key);

	/*
	 * Send back our response.
	 */
	inquiry_req.opcode = FAS_UNIT_INQUIRY;
	if (write(s, &inquiry_req, FASCMDSZ) < 0) {
		perror("write");
		close(s);
		return (-1);
	}

	if (sock_read(s, (caddr_t)&inquiry_rsp, FASCMDSZ) < 0) {
		perror("read");
		close(s);
		return(-1);
	}

	if (inquiry_rsp.opcode != FAS_UNIT_INQUIRY) {
		printf("Bad inquiry response received from "
		       "EtherLite module.\n");
		close(s);
		return(-1);
	}

	/* Consume the remainder of the FAS_UNIT_INQUIRY response. */
	if (sock_read(s, (caddr_t)str_inquiry,
		      256*inquiry_rsp.cnt_hi+inquiry_rsp.cnt_lo) < 0) {
		perror("read");
		close(s);
		return(-1);
	}

	/*
	 * Done with the challenge.  Now we can issue the reset command.
	 */
	reset_req.opcode = FAS_UNIT_RESET;
	if (write(s, &reset_req, FASCMDSZ) < 0) {
		perror("write");
		close(s);
		return(-1);
	}

	bzero((char *)&reset_rsp, FASCMDSZ);
	if (sock_read(s, (caddr_t)&reset_rsp, FASCMDSZ) < 0) {
		perror("read");
		close(s);
		return(-1);
	}

	if (reset_rsp.opcode != FAS_UNIT_RESET) {
		printf("Bad reset response received from EtherLite module.\n"
		       "You will need to upgrade the firmware in the module "
		       "to version 6.0\n"
		       "or later before you can reset it remotely.\n");
		close(s);
		return(-1);
	}

	close(s);
	return(0);
}


/*------------------------------------------------------------------*/
int sock_read(int s, caddr_t buf, int wanted) {
	int n;

	while (wanted)  {
		if ((n = read(s, buf, wanted)) <= 0) {
			return(-1);
		}
		buf += n;
		wanted -= n;
	}
	return(0);
}


/*------------------------------------------------------------------*/
/*
 *  bitmash function
 *
 *  Accepts an 6 byte challenge and returns a 3 byte key
 *  which has been bitmashed in a moderately inscrutable way.
 *  Certainly not enough security to foil a dedicated cryptographer
 *  but enough to discourage the casual cracker.
 *
 *  Those concerned with illicit access to the units should use the
 *  allowed hosts features implemented near the top of this file.
 *
 */
void bitmash(unsigned char *challenge, unsigned char *key) {
	int i, j, k, n, churns;

	/* mash the bits together */
	for( i=0, j=1, k=2, churns=0; churns < 31;
	     i = (i + 1) % 6, j = (j + i) % 6, k = (k + j) % 6, churns++ ) {
		/* XOR a pair together */
		challenge[i] ^= challenge[j];
		/* rotate one byte each time */
		n = challenge[k] << 1;
		if( n & 0x100 )
			n |= 1;
		challenge[k] = (unsigned char) n;
	}
	/* build a key */
	key[0] = challenge[0] + challenge[4];
	key[1] = challenge[2] + challenge[1];
	key[2] = challenge[3] + challenge[5] + challenge[0];

	return;
}

/* finis */
