/*****************************************************************************
 *
 * cdstty.c - set/report extended control functions for ELS units.
 *
 * Copyright 1999-2000 Digi International (www.digi.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************/

/*
 *  Options:
 *  port			        device name to use
 *  sane				resets to defaults
 *  [baud rate]				(if extended baud rate, sets extension)
 *  baud [numeric]			new form for setting baud rate
 *  loop [on/off]			enables/disables loopback
 *  errlev [numeric or mnemonic]        set error verbosity level
 *  baudext [numeric]			set baud rate extension factor
 *  cltmr [numeric secs]		close watchdog timer
 *  lcltmr[numeric secs]		local close watchdog timer
 *  outtmr [numeric secs]		wait output watchdog timer
 *  louttmr [numeric secs]		local wait output watchdog timer
 *
 *  Also optionally reports:
 *  product				ELS unit's product ID string
 *  fwver				ELS unit's firmware version
 *  targetid				ELS unit's target ID
 *
 *	Will accept multiple ports if specified on the command line with ports
 *	option.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <signal.h>
#include "ctio.h"
#include "fas.h"
#include "sts_unittab.h"
 

# define DEFAULT_NAME       "STDIN"
# define DEFAULT_FILENO     STDIN_FILENO

#include "cdutils.h"
 
#ifndef LOCAL
# define LOCAL extern
#endif
 
#include "sts_unittab.h"

#define MAX_PORTS	1024
#define dprint printf

/*
 *	macro to set up for dealing with the passed args
 */
#define init_args(argc,argv) pub_argc = argc;pub_argv = argv;arg_index = 1;


/*
 *	action function decs
 */
void port();
void sane();
void baud();
void loop();
void errlev();
void baudext();
void cltmr();
void lcltmr();
void outtmr();
void louttmr();
void displayall();
void verbose();
void verboseall();
int open_port(int port);
/*
 * option table
 */
struct optab {
	char *op;
	void (*action)();
} optab[] = {
	{"port",	port},
	{"ports",	port},
	{"sane",	sane},
	{"baudext",	baudext},
	{"baud",	baud},
	{"loop",	loop},
	{"errlev",	errlev},
	{"cltmr",	cltmr},
	{"lcltmr",	lcltmr},
	{"outtmr",	outtmr},
	{"louttmr",	louttmr},
	{"-a",		displayall},
	{"-v",		verbose},
	{"-av",		verboseall},
	{"-va",		verboseall},
	{NULL,		NULL}
};



/*
 * other protos
 */
int  main(int argc, char **argv);
void process_option(char *op);
void do_port(int port);
void apply_options(void);

void do_display(void);
char *get_eff_baud(void);

void get_port(int argc, char **argv);
void open_alarm();
void usage(void);
char *next_arg(void);

int getnum(char *s);
int isnum(char *n);
int atoi(const char *str);


/* 
 * Unix baudrate codes and associated information
 *
 * Note: FAS rates are provided simply to allow prechecking of the baudrates
 *       without making ioctl() calls.  To further this, the rate of 0 (B0) 
 *       uses FAS75 to allow it to pass this precheck on all units.
 */
struct {
	int			 rate;
	int			 code;
	int			 fas;
	char			*name;
	
} static		baudtab[] = {
	{0,	B0,	FAS75, /*fake*/	"0 (not supported)"},
	{50,	B50,	FAS50,	   "50"},
	{75,	B75,	FAS75,	   "75"},
	{110,	B110,	FAS110,	   "110"},
	{134,	B134,	FAS134,	   "134.5"},
	{150,	B150,	FAS150,	   "150"},
	{200,	B200,	FAS200,	   "200"},
	{300,	B300,	FAS300,	   "300"},
	{600,	B600,	FAS600,	   "600"},
	{1200,	B1200,	FAS1200,   "1200"},
	{1800,	B1800,	FAS1800,   "1800"},
	{2400,	B2400,	FAS2400,   "2400"},
	{4800,	B4800,	FAS4800,   "4800"},
	{9600,	B9600,	FAS9600,   "9600"},
	{19200,	B19200,	FAS19200,  "19200"},
	{38400,	B38400,	FAS38400,  "38400"},
	{57600,	B57600,	FAS57600,  "57600"},
	{115200,B115200,FAS115200, "115200"},
	{230400,B230400,FAS230400, "230400"},
	{460800,B460800,FAS460800, "460800"}
#if defined(SPARCBAUDS)
	,{76800,B76800,	FAS76800,  "76800"},
	{153600,B153600,FAS153600, "153600"},
	{307200,B307200,FAS307200, "307200"}
#endif

#if defined(SPARCBAUDS) 
# define NBAUDS  23
#else
# define NBAUDS 20
#endif
};





/*
 * public data
 */
char  *prog_name;                   /* argv[0] */
char  *tty_name = DEFAULT_NAME;     /* name of device */
int    tty_fd = DEFAULT_FILENO;     /* default file descrip */
int    pub_argc;                    /* for handling passed command line */
char **pub_argv;                    /* ditto */
int    arg_index;                   /* ditto */
char   display_default = TRUE;      /* display default amt of info */
char   display_all;                 /* display all settings */
char   verbose_flg;                 /* give longer descriptions */
char   update_ext;                  /* need to update extended params */
char   update_tty;                  /* need to update normal tty stuff */
char   baud_set = FALSE;            /* need to know if baud was manually set */
int    numports=0;		    /* number of ports to work on */
char  *portnames[MAX_PORTS];	    /* pointers to store names of ports */
struct termios ttyparms;	    /* copies of the tty and extended info */
stsextctl xc;

 

/* 
 *	store the gathered information in these variables
 */
int setsane=-1;
int baudrate=-1;
int setloop=-1;
int nonloc_timeout=-1;
int local_timeout=-1;
int nonloc_waitout=-1;
int local_waitout=-1;
int baud_ext=-1;
int level=-1;


/****************************************************************************
 *
 *	main()
 */
int main(int argc, char **argv) {

	char *op;
	int i=0;
  
#if defined(DEBUG)
	for (i=0; i < NBAUDS; i++)
		printf("%s %d %d %d %x\n", baudtab[i].name, baudtab[i].rate, 
		       baudtab[i].code, baudtab[i].fas);
#endif

	/* get setup */
	init_args(argc, argv);
	prog_name = argv[0];
	
	/* process comand line options */
	while ((op = next_arg()) != NULL)
		process_option(op);
	
	for (i=0; i <= numports; i++) 
		do_port(i);
	
	if (numports == 0)
		do_display();
	
	exit(0);
}



/*
 * process_option. scan option table for match, and if found, 
 * execute that function and return. If no match on options,
 * check for a numeric rate and store that baud rate. If 
 * neither option or baud rate found, complain.
 */
void process_option(char *op) 
{
	struct optab *o;
	int baud;
	
	o = optab;
	while (o->op != NULL) {
		if (strcasecmp(o->op,op) == 0) {
			(*o->action)();
			return;
		}
		o++;
	}
	
	/* fell thru, check for numeric baud spec */
	if (isnum(op)) {			
		baud = atoi(op);
		if (baud)
			baudrate = baud;
	}
	else {
		fprintf(stderr,"Option '%s' not recognized\n",op);
		usage();
	}
}


/*
 * optiontab procedures
 *
 */

/*
 * next_arg
 *
 * Returns next argument.  Scans each arg for '=' sign.  If found,
 * splits the arg so that next call to 'next_arg' returns the part
 * after the equals sign.  The first part is returned with the equal
 * sign replaced with the null terminator char.
 */
char *next_arg(void)
{
	char *arg;
	char *eq;
	
	if (arg_index == pub_argc) return(NULL); /* none left */
	arg = pub_argv[arg_index];
	eq = strchr(arg,'=');
	
	/* didn't find '=' so just return the arg */
	if (eq == NULL) {
		arg_index++;            /* go on to next one next time */
		return(arg);
	}
	/* found '=', do tricky thing to return arg in two parts */
	*eq = 0;                        /* split the arg */
	pub_argv[arg_index] = eq + 1;   /* point it to the char after the = */
	return(arg);                    /* note: don't increment arg_index */
}


/*
 *  port
 *
 * Look for a possibly large number of ports on the command line
 *
 */
void port(void)
{

	char  *p;
	struct optab *o;
	static int ports_scanned=FALSE;
	
	if (ports_scanned) {
		fprintf(stderr, "The ports mode cannot be specified more "
			"than once\n");
		usage();
	}
	
	ports_scanned=TRUE;
	
	portnames[numports] = p;
  
	for (numports=0 ; numports < MAX_PORTS; numports++) {
		p = next_arg();
		if (p == NULL) {
			numports--;
			return;
		}
    
		o = optab;
		while (o->op != NULL) {
			if (strcasecmp(o->op, p) == 0) {
				arg_index--;	/* reprocess this option */
				numports--;
				return;
			}
			o++;
		}
		portnames[numports] = p;
	}
	numports--;
}


/*
 * sane
 *
 * Resets things back to fairly normal defaults.
 */
void sane(void)
{
	setsane = TRUE;
	update_tty = update_ext = TRUE;
	display_default = FALSE;
}

/*
 * baud
 *
 * Used only if option form baud= is used.
 */
void baud(void) {
	baudrate = getnum("baud");
}

/*
 * loop
 */
void loop(void)
{
	char *yesno;
	
	yesno = next_arg();
	if (strcasecmp(yesno,"on") == 0) 		setloop = TRUE;
	else if (strcasecmp(yesno,"off") == 0)		setloop = FALSE;
	else if (strcmp(yesno,"1") == 0) 		setloop = TRUE;
	else if (strcmp(yesno,"0") == 0) 		setloop = FALSE;
	else {
		fprintf(stderr,"loop requires on/off or 1/0 argument\n");
		usage();
	}
	update_ext = TRUE;
	display_default = FALSE;
}

/*
 * errlev
 */
void errlev(void) 
{
	char *arg;
	
	arg = next_arg();
	if (strcasecmp(arg,"quiet") == 0)        level = STSERR_QUIET;
	else if (strcasecmp(arg,"default") == 0) level = STSERR_DEFAULT;
	else if (strcasecmp(arg,"verbose") == 0) level = STSERR_VERBOSE;
	else {
		if (!isnum(arg)) {
			fprintf(stderr,	"'errlev' expects either mnemonic "
				"argument (quiet,default,\n");
			fprintf(stderr,"verbose), or numeric.\n");
			usage();
		}
		level = atoi(arg);
	}
	level &= XC_ERRLEV;
	display_default = FALSE;
	update_ext = TRUE;
}

/*
 * baudext
 */
void baudext(void) 
{
	baud_ext = getnum("baudext");
	baud_ext <<= XC_BAUDSHIFT;
	display_default = FALSE;
	update_ext = TRUE;
}

/*
 * cltmr
 */
void cltmr(void) 
{
	nonloc_timeout = getnum("cltmr");
	display_default = FALSE;
	update_ext = TRUE;
}

/*
 * lcltmr
 */
void lcltmr(void) 
{
	local_timeout = getnum("lcltmr");
	display_default = FALSE;
	update_ext = TRUE;
}


/*
 * outtmr
 */
void outtmr(void) 
{
	nonloc_waitout = getnum("outtmr");
	display_default = FALSE;
	update_ext = TRUE;
}

/*
 * louttmr
 */
void louttmr(void) 
{
	local_waitout = getnum("louttmr");
	display_default = FALSE;
	update_ext = TRUE;
}


/*
 * displayall
 */
void displayall(void) 
{
	display_all = TRUE;
}

/*
 * verbose
 */
void verbose(void) 
{
	verbose_flg = TRUE;
}

/*
 * verboseall
 */
void verboseall(void) 
{
	display_all = verbose_flg = TRUE;
}





/****************************************************************************
 * Do the actual work on the ports 
 *
 */

/*
 * do_port()
 *
 * Called from main() as needed to process each port.  Uses information
 * gathered in process_options() to determine what to do.
 *
 */
void do_port(int port) 
{
	
	/* open the port - default to DEFAULT_FILENO */
	tty_fd = open_port(port);
	
	/* read in stuff about line */
	
	if (tcgetattr(tty_fd, &ttyparms) < 0) {
		fprintf(stderr,"Error reading TTY params for line\n");
		usage();
	}
	
	if (ioctl(tty_fd, STSGETEXTCTL, &xc) < 0) {
		fprintf(stderr,
			"STSGETEXTCTL not supported.  Not a STS/ELS port?\n");
		usage();
	}
	apply_options();		/* merge the information */ 
	
	
#if 0	
	/* rpr - only change ext when specified.   if (baud_set) */
	xc.x_flags &= ~XC_BAUDEXT;
#endif
	
	
	/* see if anything requires updating */
	if (update_ext) {
		if (ioctl(tty_fd, STSSETEXTCTL, &xc) < 0) {
			perror(NULL);
			fprintf(stderr,
				"Error updating extended control "
				"info for line\n");
			exit(1);
		}
	}
	
	if (update_tty) {
		if (tcsetattr(tty_fd, TCSANOW, &ttyparms) < 0) {
			perror(NULL);
			fprintf(stderr, "%d\n", ttyparms.c_cflag & CBAUD);
			fprintf(stderr,
				"Error updating regular TTY control info "
				"for line\n");
			exit(1);
		}
	}
	
	if (close(tty_fd)) {
		perror("Error closing port");
	}
}




/*
 * open_port()
 *
 * Open the port if one was passed in on the command line.  Return the port's
 * file descriptor.  else, pass the file descriptor for the default port
 * (STDIN or STDOUT depending on the OS).
 *
 */
int open_port(int port)
{
	char path[MAX_STR];
	
	if (portnames[port] == NULL)		/* if no port stored we  */
		return(DEFAULT_FILENO);		/* return the default    */
	
	/*
	 * grab the portname and prepend the path if the user
	 * does not specify the full pathname
	 */
	tty_name = portnames[port];
	if (portnames[port][0] != '/') {
		strcpy(path, DEV_PATH"/");
		strncat(path, portnames[port], MAX_STR - 9);
		portnames[port] = path;
		portnames[port][strlen(portnames[port])] = '\0';
	}
	
	/* try to open the port */
	signal(SIGALRM, open_alarm);
	alarm(5);
	
	tty_fd = open(portnames[port], O_RDWR | O_NDELAY);
	
	if (tty_fd < 0) {
		perror(portnames[port]);
		fprintf(stderr,"Couldn't open specified port.\n");
		exit(1);
	}
	alarm(0);
	
	/* reset to non-blocking mode */
	fcntl(tty_fd, F_SETFL, 0);
	
	/* we're all set, open succeeded */
	return(tty_fd);
}


/*
 * open_alarm
 *
 * For catching a timeout on an attempt to open the spec'd port.
 */
void open_alarm() 
{
	fprintf(stderr,"Open of port '%s' timed out.\n",tty_name);
	fprintf(stderr,"Port may be blocked in open with another process.\n");
	exit(1);
}

/*
 * apply_options()
 *
 * Apply the options gathered in process_options() to the port which is
 * currently open.
 *
 */
void apply_options(void) 
{
	char nbuf[NAMLEN + 1];
	ctunitinfo *uinfo;
	
	/*
	 *	sane defaults
	 */
	if (setsane == TRUE) {
		ttyparms.c_cflag &= ~CBAUD;
		ttyparms.c_cflag &= ~CIBAUD; /* note - no split baud support */
		ttyparms.c_cflag |= B9600;
		xc.x_flags = XC_DFLT_XFLAGS;
		xc.local_timeout = XC_DFLT_LOCAL_TIMEOUT;
		xc.nonloc_timeout = XC_DFLT_NONLOC_TIMEOUT;
		xc.local_waitout = XC_DFLT_LOCAL_WAITOUT;
		xc.nonloc_waitout = XC_DFLT_NONLOC_WAITOUT;
	}

	/*
	 * Baudrate extensions before baudrate
	 */
	if (baud_ext != -1) {
		int max_ext;
		
		strncpy(nbuf, xc.product_id, NAMLEN);
		nbuf[NAMLEN] = 0;
		uinfo = ctm_get_unitinfo(nbuf);
		max_ext = PCC_MAX_BAUD_EXT(uinfo->port_cap_code) << 
			XC_BAUDSHIFT;
		
		xc.x_flags &= ~XC_BAUDEXT;
		
		if (baud_ext > max_ext) {
			fprintf(stderr, "%s: baud extension is more "
				"than product ", tty_name);
			fprintf(stderr, "limit.  Set to limit of %d.\n", 
				PCC_MAX_BAUD_EXT(uinfo->port_cap_code));
			xc.x_flags |= (max_ext & XC_BAUDEXT);
		}
		else {
			xc.x_flags |= (baud_ext & XC_BAUDEXT);
		}
	}
  
	/*
	 * Baudrate setting code. Don't bother with extensions.
	 */
	if (baudrate != -1) {
		int i, cbaud, type;
		
		strncpy(nbuf, xc.product_id, NAMLEN);
		nbuf[NAMLEN] = 0;
		
		uinfo = ctm_get_unitinfo(nbuf);
		type = ctf_port_type(uinfo, 0);
		
		/* look for baud rate in table */
		for (i = 0; i < NBAUDS; i++) {
			if (baudtab[i].rate == baudrate)
				break;
		}
    
		/* see if we found it */
		if (i == NBAUDS) {
			fprintf(stderr,"Baudrate %u not supported\n",baudrate);
			exit(1);
		}
		
		if (ctf_check_baud(uinfo, 0, baudtab[i].fas) != OK) {
			fprintf(stderr,"Baudrate %u not supported on "
				"this product.\n", baudrate);
			exit(1);
		}
		
		
		/* found it */
		cbaud = baudtab[i].code;
		ttyparms.c_cflag &= ~CBAUD;
#if 0
		/* we don't support split baudrates under linux */
		ttyparms.c_cflag &= ~CIBAUD; 
#endif
		ttyparms.c_cflag |= cbaud;
		update_tty = TRUE;		/* make sure we get updated */
	}
	
	/*
	 * Set loop on or off as needed
	 */
	if (setloop != -1) {
		if (setloop == TRUE)
			xc.x_flags |=  XC_LOOP;
		else
			xc.x_flags &= ~XC_LOOP;
	}
  
  
	/*
	 * Set error reporting level as needed
	 */
	if (level != -1) {
		xc.x_flags &= ~XC_ERRLEV;
		xc.x_flags |= level;
	}
  
	/*
	 * cltmr
	 */
	if (nonloc_timeout != -1) {
		xc.nonloc_timeout = nonloc_timeout;
	}
  
  
	/*
	 * lcltmr
	 */
	if (local_timeout != -1) {
		xc.local_timeout = local_timeout;
	}
  
	/*
	 * outtmr
	 */
	if (nonloc_waitout != -1) {
		xc.nonloc_waitout = nonloc_waitout;
	}
  
	/*
	 * louttmr
	 */
	if (local_waitout != -1) {
		xc.local_waitout = local_waitout;
	}
	return;
}


/*
 * do_display()
 *
 * Called from main() if only did stuff to one port.  Displays the 
 * information as requested.
 *
 */
void do_display(void)
{
	
	static char *errlevs[16] = { "quiet", "default", "verbose", "3", "4", 
				     "5", "6", "7", "8", "9", "10", "11", 
				     "12", "13", "14", "15" };
	char nbuf[NAMLEN + 1];
	char vbuf[VERLEN + 1];
	
	if (display_all == 0 && display_default == 0) 
		return;
	
	/* first the more essential stuff */
	if (verbose_flg) {
		fprintf(stderr,"Port=%s\n",tty_name);
		fprintf(stderr,"Effective baud rate: %s\n", get_eff_baud());
		fprintf(stderr,"Hardware local loopback: %s\n",
			xc.x_flags & XC_LOOP ? "on" : "off");
		fprintf(stderr,"Error reporting level: %s\n",
			errlevs[xc.x_flags & XC_ERRLEV]);
	}
	else {
		fprintf(stderr,"(%s) %s loop=%s errlev=%s\n",
			tty_name,
			get_eff_baud(),
			xc.x_flags & XC_LOOP ? "on" : "off",
			errlevs[xc.x_flags & XC_ERRLEV]);
	}
  
  
	if (display_all == 0) 
		return;
	
	/* they want it all */
	strncpy(nbuf, xc.product_id, NAMLEN);    nbuf[NAMLEN] = 0;
	strncpy(vbuf, xc.fw_version, VERLEN);    vbuf[VERLEN] = 0;
	
	/* lots of words */
	if (verbose_flg) {
		fprintf(stderr,"Baud rate extension factor: %d\n",
			(xc.x_flags & XC_BAUDEXT) >> XC_BAUDSHIFT);
		fprintf(stderr,"Close watchdog timer: %d secs %s\n",
			xc.nonloc_timeout, xc.nonloc_timeout ? 
			"" : "(not enabled)" );
		fprintf(stderr,"Local close watchdog timer: %u secs %s\n",
			xc.local_timeout, xc.local_timeout ? 
			"" : "(not enabled)" );
		fprintf(stderr, "Output completion watchdog timer: "
			"%u secs %s\n",	xc.nonloc_waitout,xc.nonloc_waitout ? 
			"" : "(not enabled)" );
		fprintf(stderr,"Local output completion watchdog timer: "
			"%u secs %s\n", xc.local_waitout,xc.local_waitout ?
			"" : "(not enabled)" );
		fprintf(stderr,"Product ID: %s\nFirmware version: %s\n",
			nbuf,vbuf);
		fprintf(stderr,"target ID: %u\n", xc.scsi_id);
		
	}
	/* cool and terse */
	else {
		fprintf(stderr,"baudext=%d cltmr=%d lcltmr=%d\n",
			(xc.x_flags & XC_BAUDEXT) >> XC_BAUDSHIFT,
			xc.nonloc_timeout,xc.local_timeout);
		fprintf(stderr,"outtmr=%u louttmr=%u\n",
			xc.nonloc_waitout,xc.local_waitout);
		fprintf(stderr,"product=%s fwver=%s  targetid=%u\n",
			nbuf,vbuf,xc.scsi_id);
	}
}


/*
 * get_eff_baud
 *
 * Returns effective baud rate taking into account the baud_exten.
 */
char *get_eff_baud(void) 
{
	unsigned long cbaud;
	int baud_exten = (xc.x_flags & XC_BAUDEXT) >> XC_BAUDSHIFT;
	int i;

	cbaud = ttyparms.c_cflag & (CBAUD);

	/* look for baud rate in table */
	for (i = 0; i < NBAUDS; i++) {
		if (baudtab[i].code == cbaud)
			return (((i+baud_exten) < NBAUDS) ?
				baudtab[i+baud_exten].name :
				"out of extended range");
	}
	fprintf(stderr,"Effective baud rate out of range!\n");
	exit(1);
}




/*
 * getnum
 */
int getnum(char *s) 
{
	char *n;
	
	n = next_arg();
	if (n == NULL) {
		fprintf(stderr,"Numeric argument required for %s option\n",s);
		exit(1);
	}
	if (!isnum(n)) {
		fprintf(stderr,"Numeric argument required for %s option\n",s);
		exit(1);
	}
	return(atoi(n));
}

/*
 * isnum
 *
 * Scans string makig sure all characters are nice decimal numbers.
 * Returns TRUE if nicely formed decimal number, otherwise FALSE.
 */
int isnum(char *n) 
{
	if (*n == 0) return(FALSE);	/* make sure we got *something* */
	for ( ; *n; n++) {			/* scan */
		if (*n < '0' || *n > '9')	/* oops */
			return(FALSE);
	}
	return(TRUE);			/* fell thru, nice number */
}


/*
 * usage()
 *
 * Display the proper usage and exit
 *
 */
void usage(void) 
{
	int i;
	static char *use[] = {
#if (DEFAULT_FILENO == STDIN_FILENO)
    "Usage: cdstty [-av] [modes] [[ < ELS_ttydev]]",
#else
    "Usage: cdstty [-av] [modes] [[ > ELS_ttydev]]",
#endif
	
    "Options:",
    "-a                            display all parameters",
    "-v                            print longer parameter descriptions",
    "",
    "Modes:",
    "port                          port(s) to operate on",
    "sane                          restore default settings",
    "[numeric baud rate]           (if extended baud rate, sets extension)",
    "loop [on/off]                 enables/disables loopback",
    "errlev [numeric or mnemonic]  set error verbosity level",
    "baudext [numeric]             set baud rate extension factor",
    "cltmr [numeric secs]          close watchdog timer",
    "lcltmr[numeric secs]          local close watchdog timer",
    "outtmr [numeric secs]         wait output watchdog timer",
    "louttmr [numeric secs]        local wait output watchdog timer",
    "",
    NULL };
	
	for (i = 0; use[i]; i++)
		fprintf(stderr,"%s\n",use[i]);

	exit(1);
}






